/* TW523 X10 powerline transmission driver
 * Copyright (C) 2000  Mark Aikens <marka@wpi.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/init.h>

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/parport.h>
#include <linux/mm.h>

#include <asm/uaccess.h>

#include "tw.h"


/* Global variables */
static int num_devices=0;		/* total number of tw's found */
static unsigned long twdelay=TWDELAY;	/* calibrated delay value */

#ifdef MODULE
static int debug=0;
static int parport[MAX_DEVICES]={[0 ... MAX_DEVICES-1] = -1};

MODULE_PARM(debug, "i");
MODULE_PARM(parport, "1-" __MODULE_STRING(MAX_DEVICES) "i");
#else
#warning Not compiling as a module is completely UNTESTED
static int debug __initdata = 0;
static int parport[MAX_DEVICES] __initdata = {[0 ... MAX_DEVICES-1] = -1};
#endif

static struct tw_config tw_dev[MAX_DEVICES];

static struct file_operations tw_fops = {
	read:		tw_read,
	write:		tw_write,
	open:		tw_open,
	release:	tw_release
};

#define DEBUG(format, args...) if (debug>0) printk(KERN_DEBUG DEV_NAME ": " \
	format, ## args)
#define LOG(format, args...) printk(KERN_NOTICE DEV_NAME ": " format, ## args)
#define ERROR(format, args...) printk(KERN_ERR DEV_NAME ": " format, ## args)

/* X10 start code */
static char *x10_start="1110";

/* X10 housecodes (A-P) in complemented form */
static char *x10_house[16]={
	"01101001","10101001","01011001","10011001",
	"01010110","10010110","01100110","10100110",
	"01101010","10101010","01011010","10011010",
	"01010101","10010101","01100101","10100101"};

/* X10 keycodes in complemented form */
static char *x10_key[32]={
	"0110100101",	/* 1 */
	"1010100101",	/* 2 */
	"0101100101",	/* 3 */
	"1001100101",	/* 4 */
	"0101011001",	/* 5 */
	"1001011001",	/* 6 */
	"0110011001",	/* 7 */
	"1010011001",	/* 8 */
	"0110101001",	/* 9 */
	"1010101001",	/* 10 */
	"0101101001",	/* 11 */
	"1001101001",	/* 12 */
	"0101010101",	/* 13 */
	"1001010101",	/* 14 */
	"0110010101",	/* 15 */
	"1010010101",	/* 16 */
	"0101010110",	/* All Units Off */
	"0101011010",	/* All Units On */
	"0101100110",	/* On */
	"0101101010",	/* Off */
	"0110010110",	/* Dim */
	"0110011010",	/* Bright */
	"0110100110",	/* All Lights Off */
	"0110101010",	/* Extended Code */
	"1001010110",	/* Hail Request */
	"1001011010",	/* Hail Acknowledge */
	"1001100110",	/* Preset Dim 0 */
	"1001101010",	/* Preset Dim 1 */
	"1010010101",	/* Extended Data */
	"1010011010",	/* Status = on */
	"1010100110",	/* Status = off */
	"1010101010"	/* Status request */
};

/* Function: twdelay25
 * Description:
 *   A tight loop to delay 25us based on the twdelay constant calculated
 *   during module loading.
 */
static inline void twdelay25(void) {
	unsigned int i;
	for(i=0; i<twdelay; i++);
}

/* Function: twdelay1ms
 * Description:
 *   Delay 1ms.
 */
static inline void twdelay1ms(void) {
	int i;
	for(i=0; i<40; i++) twdelay25();
}

/* Function: wait_for_zero
 * Description:
 *   Wait for the zero crossing from the TW523 module. We wait up to 250ms
 *   for the zero crossing.
 * Parameters:
 *   p - The parport to which the TW523 module is connected.
 * Returns:
 *   The number of microseconds waited until the zero crossing or
 *   -1 if a zero crossing was not found within 250ms.
 *
 *   A zero crossing on the power line is signalled by a bit transition
 *   on the zero crossing output from the TW523.
 */
static int wait_for_zero(struct parport *p) {
	int i, old, new;

	old = parport_read_status(p) & 0x20;
	for(i=0; i<10000; i++) {
		new = parport_read_status(p) & 0x20;
		if(new != old) return(i*25);
		twdelay25();
	}
	return -1;
}

/* Function: tw_xlat_str
 * Description:
 *   Immediatly transmits a bit string on the power line. The initial zero
 *   crossing should be found before this function is called.
 * Parameters:
 *   s - A pointer to a strings representing the bit pattern to send.
 *   p - The parport, where TW523 is located, to send the bit pattern on.
 */
static inline int tw_xlat_str(char *s, struct parport *p) {
	int i;

	for(i=0; s[i]!=0; i++) {
		if(s[i] == '1') parport_write_data(p, 0xff);
		twdelay1ms();
		parport_write_data(p, 0x00);
		if(wait_for_zero(p) < 0) return -1;
	}
	return 0;
}

ssize_t tw_read(struct file *file, char *buf, size_t len, loff_t *ppos) {
	int minor=MINOR(file->f_dentry->d_inode->i_rdev);

	DEBUG("read->%d bytes, minor: %d\n", len, minor);
	return -ENOSYS;
}

ssize_t tw_write(struct file *file, const char *buf, size_t len, loff_t *ppos) {
	int i, minor=MINOR(file->f_dentry->d_inode->i_rdev);
	u_char b[4];
	u_short house, key, count=0;
	
	DEBUG("write->%d bytes, minor %d\n", len, minor);
	if((len < 2) || (len > 3)) return -EINVAL;

	if(copy_from_user(&b, buf, len)) return -EFAULT;
	house = b[0] & 0x0f;
	key = b[1] & 0x1f;
	if(len == 3) count = b[2];
	if(count == 0) count = 2;
	DEBUG("write->data: house=%d key=%d count=%d\n",
	  (int)house, (int)key, (int)count);

	/* wait 3 full powerline cycles (six zero crossings) */
	for(i=0;i<6;i++) {
		if(wait_for_zero(tw_dev[minor].pport)<0) {
			ERROR("unable to find the start zero crossing on " \
			  "tw%d\n", minor);
			return -EIO;
		}
	}
	for(;count;count--) {
		if(tw_xlat_str(x10_start,tw_dev[minor].pport) != 0) {
			ERROR("aborted start code transmission on tw%d\n",
			  minor);
			return -EIO;
		}
		if(tw_xlat_str(x10_house[house],tw_dev[minor].pport)!=0) {
			ERROR("aborted house code transmission on tw%d\n",
			  minor);
			return -EIO;
		}
		if(tw_xlat_str(x10_key[key],tw_dev[minor].pport)!=0) {
			ERROR("aborted key code transmission on tw%d\n", minor);
			return -EIO;
		}
	}
	return(len);
}

static int tw_open(struct inode *inode, struct file *file) {
	int minor=MINOR(inode->i_rdev);

	DEBUG("open, minor: %d\n", minor);
	if(minor >= num_devices) return -ENXIO;
	if(atomic_inc_and_test_greater_zero(&tw_dev[minor].opencount)) {
		atomic_dec(&tw_dev[minor].opencount);
		DEBUG("open on device in use, minor: %d\n",minor);
		return -EBUSY;
	}
	MOD_INC_USE_COUNT;
	return 0;
}


static int tw_release(struct inode *inode, struct file *file) {
	int minor=MINOR(inode->i_rdev);

	DEBUG("release, minor: %d\n", minor);
	atomic_dec(&tw_dev[minor].opencount);
	MOD_DEC_USE_COUNT;
	return 0;
}

/* Function: tw_init
 * Description:
 *   Probe for a TW523 module on a parallel port.
 * Parameters:
 *   minor - The minor device number of the TW523 device we are probing for.
 *   p - A parport structure of the parallel port to probe on.
 * Returns:
 *   0 if a TW523 was found on the parallel port, otherwise an error number
 *   from linux/errno.h
 */
int tw_init(int minor, struct parport *p) {
	unsigned int delay;
	int i;

	if(parport_in_use(p)) {
		DEBUG("parport%d is in use by another device\n", p->number);
		return -EIO;
	}

	tw_dev[minor].pport = p;
	atomic_dec(&tw_dev[minor].opencount);	/* opencount = -1 */
	tw_dev[minor].pdev = parport_register_device(p, DEV_NAME, NULL,
	  NULL, NULL, PARPORT_DEV_EXCL, NULL);
	if(tw_dev[minor].pdev == NULL) return -EIO;

	parport_claim(tw_dev[minor].pdev);
	parport_write_data(tw_dev[minor].pport, 0);
	parport_write_control(tw_dev[minor].pport, 0);

	DEBUG("checking on parport%d\n", tw_dev[minor].pport->number);
	/* wait for initial zero crossing */
	if(wait_for_zero(tw_dev[minor].pport)<0) {
		DEBUG("not found on parport%d\n", tw_dev[minor].pport->number);
		parport_release(tw_dev[minor].pdev);
		parport_unregister_device(tw_dev[minor].pdev);
		return -EIO;
	}

	if(minor == 0) {			/* only calibrate once */
		for(i=0; i<10; i++) {
			delay = wait_for_zero(tw_dev[minor].pport);
			if(delay > 0) twdelay = (twdelay*delay)/HALFCYCLE;
		}
		DEBUG("twdelay=%lu\n", twdelay);
	}
	if(wait_for_zero(tw_dev[minor].pport) < 0) {
		parport_release(tw_dev[minor].pdev);
		parport_unregister_device(tw_dev[minor].pdev);
		return -EIO;
	}
	DEBUG("device found on parport%d\n", tw_dev[minor].pport->number);
	return 0;
}

#ifdef MODULE
int init_module(void) {
	if(debug != 1) debug = 0;
	LOG("tw powerline interface kernel module v%s loaded.\n", TW_VERSION);
	DEBUG("Debugging output enabled.\n");

	if(register_chrdev(TW_MAJOR, DEV_NAME, &tw_fops)) {
		LOG("Cannot register to major device: %d\n", TW_MAJOR);
		return -EIO;
	}
	DEBUG("Registered to major device %d.\n", TW_MAJOR);

	if(parport[0] != -1) {		/* module parameter was specified */
		struct parport *pp;
		int i;
		for(i=0; (parport[i]!=-1)&&(i<MAX_DEVICES); i++) {
			for(pp=parport_enumerate(); pp; pp=pp->next) {
				if(pp->number == parport[i]) {
					if(tw_init(num_devices,pp) == 0)
						num_devices++;
				}
			}
		}
	} else {			/* enumerate through all parports */
		struct parport *pp;
		for(pp=parport_enumerate(); pp; pp=pp->next) {
			if(tw_init(num_devices,pp) == 0) num_devices++;
		}
	}
	if(num_devices == 0) {
		LOG("No devices found. Not loading module.\n");
		unregister_chrdev(TW_MAJOR, DEV_NAME);
		return -ENODEV;
	}
	DEBUG("Number of tw's configured: %d\n", num_devices);
	return 0;
}


void cleanup_module(void) {
	int i;

	unregister_chrdev(TW_MAJOR, DEV_NAME);
	for(i=0;i<num_devices;i++) {
		parport_release(tw_dev[i].pdev);
		parport_unregister_device(tw_dev[i].pdev);
		DEBUG("Released parport%d\n",tw_dev[i].pport->number);
	}
	LOG("Module unloaded\n");
}
#endif
